/*
** Module   : HandleOps.java
** Abstract : Defines APIs to implement handle-specific operations.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130215 Added support for DELETE WIDGET. All deletes are done at the resource, as they
**                  need to implement Deletable.
** 003 CA  20130221 getType was renamed to getResourceType.
** 004 GES 20130302 Added CAN-QUERY and CAN-SET builtin function support.
** 005 CA  20130529 Removed explicit handling of procedure deletion, now all is done in a generic
**                  approach.
** 006 CA  20130918 Resource cleanup must be done only if the resource deleted. 
** 007 EVL 20130911 Adding implementation for CAN-QUERY/CAN-SET functions.
** 008 EVL 20131009 Fix for getting the attribute getters and setters.
*/

package com.goldencode.p2j.util;

import com.goldencode.proxy.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * Provides implementation for these handle-specific operations: 
 * <ul>
 *    <li>DELETE OBJECT</li>
 *    <li>DELETE PROCEDURE</li>
 *    <li>CAN-QUERY() built-in function</li>
 *    <li>CAN-SET() built-in function</li>
 * </ul>
 * 
 * Current implementation provides support only for SESSION, FILE-INFO, 
 * LAST-EVENT, ERROR-STATUS and procedure handles.
 */
public class HandleOps
{
   /**
    * Deletes the given handles. If multiple handles refer the same resource, then the resource is
    * deleted only once (the first time is encountered).
    * <p>
    * In cases when the handle is a system handle, an {@link #invalidDelete()}
    * is called (to raise an error).
    * 
    * @param    handles
    *           The list of handles to be deleted.
    */
   public static void delete(handle... handles)
   {
      Set<Object> unique = new HashSet<Object>();
      for (handle h : handles)
      {
         if (h == null)
         {
            invalidDelete();
            break;
         }

         Object referent = h.get();
         if (unique.contains(referent))
            continue;
         
         unique.add(referent);
         
         if (!h._isValid())
         {
            // this is shown for all invalid handles.
            String msg = "Invalid handle.  Not initialized or points to a deleted object";
            ErrorManager.displayError(3135, msg);
            continue;
         }
         
         // DELETE WIDGET can't be protected with no error, so if something fails here, the error
         // will propagate
         delete(h);
      }
   }

   /**
    * Deletes the given handle. The current implementation allows only
    * procedure handles to be deleted.
    * <p>
    * In cases when the handle is a system handle, an {@link #invalidDelete()}
    * is called (to raise an error).
    * 
    * @param    h
    *           The handle to be deleted.
    */
   public static void delete(handle h)
   {
      if (h != null && h._isValid())
      {
         Object referent = h.get();
         if (h.equals(SessionUtils.asHandle())  ||
             h.equals(FileSystemOps.asHandle()) ||
             h.equals(KeyReader.asHandle())     ||
             h.equals(ErrorManager.asHandle())  ||
             ProcedureManager.systemHandle(referent))
         {
            invalidDelete();
            return;
         }
         
         WrappedResource resource = h.getResource();
         if (resource instanceof Deletable)
         {
            ((Deletable) resource).delete();
            // do not perform any cleanup here. delete() doesn't ensure the resource was deleted
            // (an error might be raised, with NO-ERROR clause present), so cleanup must be done
            // either in HandleChain.delete (if the resource extends HandleChain) or in resource's
            // delete implementation, if the resource can be deleted
            return;
         }
         
         throw new RuntimeException(
            "Delete implementation not available for handle " + 
            handle.unwrapType(h).getResourceType().toStringMessage());
      }
      
      invalidDelete();
   }

   /**
    * Reports if the given attribute can be read from the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-QUERY()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports reading the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canQuery(handle h, character attr)
   {
      if (attr == null || attr.isUnknown())
      {
         return new logical();
      }
      
      return canQuery(h, attr.toStringMessage());
   }
   
   /**
    * Reports if the given attribute can be read from the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-QUERY()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports reading the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canQuery(handle h, String attr)
   {
      if (h == null || h.isUnknown())
      {
         return new logical();
      }
      
      LegacyAttribute legacyAnnotation = getLegacyAttribute(h, attr, false);
      
      if (isHandleAttribute(attr) ||
          (legacyAnnotation != null && !legacyAnnotation.ignore()))
      {
         return new logical(true);
      }

      return new logical(false);
   }
   
   /**
    * Reports if the given attribute can be written to the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-SET()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports writing the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canSet(handle h, character attr)
   {
      if (attr == null || attr.isUnknown())
      {
         return new logical();
      }
      
      return canSet(h, attr.toStringMessage());
   }
   
   /**
    * Reports if the given attribute can be written to the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-SET()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports writing the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canSet(handle h, String attr)
   {
      if (h == null || h.isUnknown())
      {
         return new logical();
      }
      
      LegacyAttribute legacyAnnotation = getLegacyAttribute(h, attr, true);
      
      if (!isHandleAttribute(attr) &&
          (legacyAnnotation != null && !legacyAnnotation.ignore()))
      {
         return new logical(true);
      }

      return new logical(false);
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given method  and given legacy
    * attribute name.
    * 
    * @param    meth_check
    *           The method to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given class
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttributeInMethod(Method meth_check,
                                                             String attr,
                                                             boolean setter)
   {
      LegacyAttribute legacyAttrAnno = meth_check.getAnnotation(LegacyAttribute.class);
      // Check annotation for name match(ignoring letter case)
      if (legacyAttrAnno != null && legacyAttrAnno.name().equalsIgnoreCase(attr))
      {
         boolean settable = legacyAttrAnno.setter(); 
         if ((!setter && !settable) || (setter && settable))
         {
            return legacyAttrAnno;
         }
      }
      
      return null;
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given implementation or interface
    * class object and given legacy attribute name.
    * 
    * @param    class_check
    *           The interface or implementation class object to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given class
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttributeInClass(Class class_check,
                                                            String attr,
                                                            boolean setter)
   {
      // For each interface implementation get the methods and check each one
      Method[] methods = class_check.getMethods();
      if (methods != null)
      {
         for (Method method : methods)
         {
            // Check every method for match
            LegacyAttribute legacyAttrAnno = getLegacyAttributeInMethod(method, attr, setter);
            // Check if we found acceptable annotation
            if (legacyAttrAnno != null)
            {
               return legacyAttrAnno;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given static class object and
    * given legacy attribute name.
    * 
    * @param    obj_check
    *           The class object to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given class
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttributeInStatic(Object obj_check,
                                                             String attr,
                                                             boolean setter)
   {
      // For each interface implementation get the methods and check each one
      List<Method> staticMethods = StaticProxy.getStaticMethods(obj_check);
      if (staticMethods != null)
      {
         for (Method method : staticMethods)
         {
            // Check every method for match
            LegacyAttribute legacyAttrAnno = getLegacyAttributeInMethod(method, attr, setter);
            // Check if we found acceptable annotation
            if (legacyAttrAnno != null)
            {
               return legacyAttrAnno;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given handle object and given
    * legacy attribute name.
    * <p>
    * The search algorithm for this method uses reflection to read annotations from various
    * sources in the following precedence order:
    * <ol>
    *   <li> The proxy static methods of the resource itself if exist.
    *   <li> The class of the resource itself.
    *   <li> The superclass of the resource or the superclass of the superclass... the entire
    *        parent hierarchy is walked from the resource to the root.  The first match found ends
    *        the upward search.
    *   <li> All interfaces of the resource itself.
    *   <li> All interfaces of the superclass of the resource or the superclass of the
    *        superclass... the entire parent hierarchy is walked from the resource to the root.
    *        The first superclass with an interface that has a matching attribute will end the
    *        upward search.
    * </ol>
    * <p>
    * Where possible, we put annotations on the interface, not the class for the resource.  But
    * there are some attributes where the configuration is not universal for all resources
    * implementing the same attribute.  For example, the following attributes are readable for
    * certain resources even though for most other resources the same attribute is writable:
    * <ul>
    *   <li> LABEL for LAST-EVENT
    *   <li> MODIFIED for browse columns
    *   <li> COLUMN for browse cells, field groups, and LAST-EVENT
    *   <li> NO-VALIDATE for both static and dynamic browses (it is writeable only for dynamic
    *        browses)
    * </ul>
    * <p>
    * Keeping the above precedence order is critical to allow attribute configurations of
    * more-specific resources to properly override the more general configuration in the parent
    * hierarchy or the interface hierarchy.
    * 
    * @param    h
    *           The handle object to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given handle
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttribute(handle h, String attr, boolean setter)
   {
      LegacyAttribute legacyAnnotationRes = null;
      Object objFromHandle = h.get();
      
      if (objFromHandle != null)
      {
         // Checking the static proxy first
         legacyAnnotationRes = getLegacyAttributeInStatic(objFromHandle, attr, setter);
         // Found result in static methods
         if (legacyAnnotationRes != null)
         {
            return legacyAnnotationRes;
         }
         
         // Getting the base class
         Class current_base = objFromHandle.getClass(); 
         // Getting implementation methods from current class first
         while (current_base != null)
         {
            // For each implementation get the methods and check each one
            legacyAnnotationRes = getLegacyAttributeInClass(current_base, attr, setter);
            // Found result in implementation
            if (legacyAnnotationRes != null)
            {
               return legacyAnnotationRes;
            }
            // If we are here the annotation is not found, try superclass if exists
            current_base = current_base.getSuperclass(); 
         }

         // Restore the search from the base class
         current_base = objFromHandle.getClass();
         // Getting iterfaces from current class
         while (current_base != null)
         {
            Class[] classes = current_base.getInterfaces();
            if (classes != null)
            {
               for (Class iface_curr : classes)
               {
                  // For each interface implementation get the methods and check each one
                  legacyAnnotationRes = getLegacyAttributeInClass(iface_curr, attr, setter);
                  // Found result in implementation
                  if (legacyAnnotationRes != null)
                  {
                     return legacyAnnotationRes;
                  }
               }
            }
            // If we are here the annotation is not found, try superclass if exists
            current_base = current_base.getSuperclass(); 
         }
      }
      
      return null;
   }
   
   /**
    * Checks if the given attribute name is HANDLE.  This is used to bypass reflection based code
    * to find out if the attribute can be set or queried.  We know this is read-only and always
    * exists for valid handle we use in CAN-SET and CAN-QUERY calls.
    * 
    * @param    attrName
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the attribute name is HANDLE, <code>false</code> otherwise.
    */
   private static boolean isHandleAttribute(String attrName)
   {
      return attrName.equalsIgnoreCase("HANDLE");
   }
   
   /**
    * Invoked when the handle is invalid or a system (non-procedure) handle. 
    * Shows appropriate error messages and raises an error. 
    * Calls {@link #invalidDelete(handle) invalidDelete(null)}.
    */
   private static void invalidDelete()
   {
      invalidDelete(null);
   }

   /**
    * Invoked when the handle is invalid or a system (non-procedure) handle. 
    * Shows appropriate error messages and raises an error.
    * 
    * @param    h
    *           The handle for which the delete failed. May be 
    *           <code>null</code> if it was decide this is a non-procedure 
    *           handle.
    */
   private static void invalidDelete(handle h)
   {
      boolean procedure = (h != null && h._isValid() && ProcedureManager.isProcedure(h));

      final String msg1 = "Invalid or inappropriate handle value given to " +
                          "DELETE OBJECT or DELETE PROCEDURE statement";
      final int id1 = 5425;

      final String msg2 = "Invalid or inappropriate handle value given to " +
                          "DELETE PROCEDURE statement. Procedure '%s':%d";
      final int id2 = 1428;
      // this can't be computed, is the legacy line num from the 4GL file.
      final int linenum = 0;
      
      int id = -1;
      String msg = null;
      
      if (procedure)
      {
         character pname = handle.unwrapProcedure(h).getFileName();
         msg = String.format(msg2, pname.toStringMessage(), linenum);
         id = id2;
      }
      else
      {
         msg = msg1;
         id = id1;
      }

      ErrorManager.recordOrThrowError(id, msg, false);
   }
}
